from restfulpy.orm.metadata import FieldInfo


def test_to_json_fieldinfo():
    age = FieldInfo(
        type_=(str, 400),
        pattern=('\\d+', 400),
        max_length=(2, 400),
        min_length=(1, 400),
        minimum=(1, 400),
        maximum=(99, 400),
        readonly=(True, 400),
        protected=(True, 400),
        not_none=(True, 400),
        required=(True, 400),
    )
    age_json_field_info = age.to_json()

    assert age_json_field_info['type'] == 'str'
    assert age_json_field_info['pattern'] == '\\d+'
    assert age_json_field_info['maxLength'] == 2
    assert age_json_field_info['minLength'] == 1
    assert age_json_field_info['minimum'] == 1
    assert age_json_field_info['maximum'] == 99
    assert age_json_field_info['readonly'] is True
    assert age_json_field_info['protected'] is True
    assert age_json_field_info['notNone'] is True
    assert age_json_field_info['required'] is True



from sqlalchemy import Unicode, Integer

from restfulpy.orm import DeclarativeBase, Field, ModifiedMixin


class ModificationCheckingModel(ModifiedMixin, DeclarativeBase):
    __tablename__ = 'modification_checking_model'
    __exclude__ = {'age'}

    title = Field(Unicode(50), primary_key=True)
    age = Field(Integer)


class ModificationExcludelessModel(ModifiedMixin, DeclarativeBase):
    __tablename__ = 'modification_checking_excludeless_model'

    title = Field(Unicode(50), primary_key=True)
    age = Field(Integer)


def test_modified_mixin(db):
    session = db()

    instance = ModificationCheckingModel(
        title='test title',
        age=1,
    )
    session.add(instance)

    excludeless_instance = ModificationExcludelessModel(
        title='test title',
        age=1,
    )
    session.add(excludeless_instance)
    session.commit()

    assert instance.modified_at is None
    assert instance.created_at is not None
    assert instance.last_modification_time == instance.created_at

    instance = session.query(ModificationCheckingModel).one()
    assert instance.modified_at is None
    assert instance.created_at is not None
    assert instance.last_modification_time == instance.created_at

    instance.age = 2
    session.commit()
    assert instance.modified_at is None
    assert instance.created_at is not None
    assert instance.last_modification_time == instance.created_at

    instance.title = 'Edited title'
    session.commit()
    assert instance.modified_at is not None
    assert instance.created_at is not None
    assert instance.last_modification_time == instance.modified_at

    instance = session.query(ModificationCheckingModel).one()
    assert instance.modified_at is not None
    assert instance.created_at is not None
    assert instance.last_modification_time == instance.modified_at

    excludeless_instance.age = 3
    session.commit()
    assert excludeless_instance.modified_at is not None
    assert excludeless_instance.created_at is not None
    assert excludeless_instance.last_modification_time == \
        excludeless_instance.modified_at




from ..utils import to_camel_case


class FieldInfo:
    def __init__(self, type_=str, max_length=None, min_length=None,
                  minimum=None, maximum=None, default=None, readonly=False,
                  pattern=None, pattern_description=None, protected=False,
                  not_none=None, required=None):
        self.type_ = type_
        self.pattern = pattern
        self.pattern_description = pattern_description
        self.max_length = max_length
        self.min_length = min_length
        self.minimum = minimum
        self.maximum = maximum
        self.readonly = readonly
        self.protected = protected
        self.not_none = not_none
        self.required = required
        self.default = default

    def to_json(self):
        result = {}

        for k,v in vars(self).items():
            json_key = to_camel_case(k)
            json_value = v[0] if isinstance(v, tuple) else v
            if isinstance(json_value, type):
                json_value = json_value.__name__

            result[json_key] = json_value

        result['type'] = result.pop('type_')
        return result

    def __copy__(self):
        new_one = type(self)()
        new_one.__dict__.update(self.__dict__)
        return new_one

    def to_dict(self):
        return {
            k: v for k, v in self.__dict__.items() if not k.startswith('_')
        }


class MetadataField(FieldInfo):
    def __init__(self, name, key, primary_key=False, label=None,
                 watermark=None, message=None, example=None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.key = key[1:] if key.startswith('_') else key
        self.primary_key = primary_key
        self.label = label
        self.watermark = watermark
        self.example = example
        self.message = message

    @classmethod
    def from_column(cls, c, info=None):
        if not info:
            info = c.info
        json_name = info.get('json') or to_camel_case(c.key)
        result = []

        key = c.key

        if hasattr(c, 'default') \
                and c.default \
                and hasattr(c.default, 'is_scalar'):
            default = c.default.arg if c.default.is_scalar else None
        else:
            default = None

        result.append(cls(
            json_name,
            key,
            default=default,
            type_=info.get('type'),
            not_none=info.get('not_none'),
            required=info.get('required'),
            pattern=info.get('pattern'),
            pattern_description=info.get('pattern_description'),
            max_length=info.get('max_length'),
            min_length=info.get('min_length'),
            minimum=info.get('minimum'),
            maximum=info.get('maximum'),
            watermark=info.get('watermark', None),
            label=info.get('label', None),
            example=info.get('example', None),
            primary_key=hasattr(c.expression, 'primary_key') \
                and c.expression.primary_key,
            readonly=info.get('readonly', False),
            protected=info.get('protected', False),
            message=info.get('message')
        ))

        return result

    def to_json(self):
        result = super().to_json()
        result.update(
            name=self.name,
            example=self.example,
            watermark=self.watermark,
            label=self.label,
            message=self.message,
            key=self.key,
            primaryKey=self.primary_key,
        )
        return result

